/*
 *  BlobDetector.cpp
 *  kinectBasic
 *
 *  Created by Spencer Salazar on 5/15/11.
 *  Copyright 2011 Spencer Salazar. All rights reserved.
 *
 */

#include "BlobDetector.h"



#define MAX_NUMOF_BLOBS 256
#define MAX_OSC_BLOBS (10)




static void fCvThreadFun(BlobDetector *bd)
{
    bd->cvThread();
}


static int qsort_blobarea_comparator( const void * elem1, const void * elem2 )
{
	float a1 = fabs( cvContourArea(*((CvSeq **)elem1)) );
	float a2 = fabs( cvContourArea(*((CvSeq **)elem2)) );
	
	return a1>a2 ? -1 : 1;
}


BlobDetector::BlobDetector()
{
    mCvBuffer = new CircularBuffer<Surface>(30);
    networkManager = NetworkManager("localhost", "6449");
}


void BlobDetector::start()
{
    mCvThread = new boost::thread(fCvThreadFun,this);
}


void BlobDetector::nextFrame(Surface frame)
{
    mCvBuffer->put(frame);
}


void BlobDetector::cvThread()
{
    cvThreadDoExit = false;
    Surface frame;
    
    IplImage * bwImage[3];
    
    bwImage[0] = cvCreateImage(cvSize(getWindowWidth(), getWindowHeight()), 
                               IPL_DEPTH_8U, 1);
    bwImage[1] = cvCreateImage(cvSize(getWindowWidth(), getWindowHeight()), 
                               IPL_DEPTH_8U, 1);
    bwImage[2] = cvCreateImage(cvSize(getWindowWidth(), getWindowHeight()), 
                               IPL_DEPTH_8U, 1);
    
    // memory storage for analysis
    CvMemStorage * storage = cvCreateMemStorage();
    // to calculate gravity center, area, etc…
    // TODO: does this really need to be dynamically allocated? -spencer
    CvMoments moments;
    // the valid contour list
    CvSeq * contours[ MAX_NUMOF_BLOBS ];
    // contour index that will walk through starting at 0 address
    CvSeq * index = 0;
    // the number of coontours found
    int count = 0;
    
    while(!cvThreadDoExit)
    {
        while(mCvBuffer->get(frame))
        {
            // copy the RGB into 1-channel black and white image
            // the image is filtered to bias towards pure-red pixels
            int threshold = 25;
            Surface::ConstIter iter = frame.getIter();
            while(iter.line())
            {
                while(iter.pixel())
                {
                    int desti = (iter.x()*bwImage[0]->nChannels)+(iter.y()*bwImage[0]->widthStep);
                    
                    // red blobs
                    if((iter.r() - iter.g()/2 - iter.b()/2) > threshold)
                        bwImage[0]->imageData[desti] = 1;
                    else
                        bwImage[0]->imageData[desti] = 0;
                    // green blobs
                    if((iter.g() - iter.r()/2 - iter.b()/2) > threshold)
                        bwImage[1]->imageData[desti] = 1;
                    else
                        bwImage[1]->imageData[desti] = 0;
                    // blue blobs
                    if((iter.b() - iter.g()/2 - iter.r()/2) > threshold)
                        bwImage[2]->imageData[desti] = 1;
                    else
                        bwImage[2]->imageData[desti] = 0;
                }
            }
            
            for(int j = 0; j < 3; j++)
            {
                // blob detect that shiz
                cvResetImageROI(bwImage[j]);
                index = 0;
                
                // find all contours
                cvFindContours(bwImage[j], storage, &index, sizeof(CvContour), CV_RETR_LIST);
                
                float minArea = 25;
                float maxArea = getWindowWidth()*getWindowHeight();
                int maxBlobs = 10;
                count = 0;
                
                // list valid blob's contour : included in the miminum and maximum area defined by the user
                // to remove noises or biggest blob
                while( index!=NULL && count<=MAX_NUMOF_BLOBS )
                {
                    float area = fabs( cvContourArea(index) );
                    if ( area>=minArea && area<=maxArea ) contours[ count++ ] = index;
                    index = index->h_next;
                }
                
                // reorder contours from the largest to the smallest 
                if ( count > 0 ) qsort( contours, count, sizeof(CvSeq*), qsort_blobarea_comparator );
                
                // limit the number of blob elements in list
                // should be limited after Sorting elements in array to retrieve all biggest blobs
                count = MIN( count, maxBlobs );
                
                vector<Blob> blobs;
                
                for( int i=0; i<count; i++ )
                {
                    // calculates all moments up to third order of a polygon or rasterized shape
                    cvMoments( contours[i], &moments );
                    
                    // set the maximum number of blob's points definition
                    //int max = MIN( maxVertices, contours[i]->total );
                    
                    Blob b;
                    
                    // retrieve CV blob's properties (area, arc length, bounding rectangle, gravity center)
                    b.area	 = cvContourArea( contours[i] );
                    b.length = cvArcLength( contours[i] );
                    CvRect rect	 = cvBoundingRect( contours[i] );
                    b.x = rect.x; b.y = rect.y;
                    b.width = rect.width; b.height = rect.height;
                    b.centerX = (int) ( moments.m10/moments.m00 );
                    b.centerY = (int) ( moments.m01/moments.m00 );
                    
                    blobs.push_back(b);
                }
                
                networkManager.sendBlobs(j, blobs);
                
                cvClearMemStorage(storage);
            }
        }
        
        usleep(1000);
    }
    
    cvReleaseMemStorage( &storage );
    cvReleaseImage(&bwImage[0]);
    cvReleaseImage(&bwImage[1]);
    cvReleaseImage(&bwImage[2]);
    bwImage[0] = bwImage[1] = bwImage[2] = 0;
}


NetworkManager::NetworkManager(const char *host, const char *port) :
mAddress(NULL),
mCurrentBlobGroup(0)
{
    if(host != NULL && port != NULL)
    {
        mAddress = lo_address_new(host, port);
    }
}


void NetworkManager::sendBlobs(int type, vector<Blob> & blobs)
{
    if(mAddress)
    {
        lo_message msg = lo_message_new();
        
        lo_message_add(msg, "ii", type, blobs.size());
        
        for(int i = 0; i < MAX_OSC_BLOBS && i < blobs.size(); i++)
        {
            const Blob &b = blobs[i];
            lo_message_add(msg, "iiffiiii", 
                           b.centerX, b.centerY,
                           b.area, b.length,
                           b.x, b.y, b.width, b.height);
        }
        
        // send exactly 10 
        if(blobs.size() < MAX_OSC_BLOBS)
        {
            Blob b;
            memset(&b, 0, sizeof(Blob));
            for(int i = blobs.size(); i < MAX_OSC_BLOBS; i++)
            {
                lo_message_add(msg, "iiffiiii", 
                               b.centerX, b.centerY,
                               b.area, b.length,
                               b.x, b.y, b.width, b.height);
            }
        }
        
        lo_send_message(mAddress, "/blobs", msg);
        
        lo_message_free(msg);
    }
}



