
#include "VideoDelay.h"
#include "VideoRecording.h"
#include "BlobDetector.h"

#include "Resources.h"

#include "CircularBuffer.h"


using namespace ci;
using namespace ci::app;
using namespace std;


const int NUM_RECORDINGS = 10;
const int NUM_VIDEO_DELAYS = 0;
const float VIDEO_DELAYS[] = { 1, 2, 3, 5, 10, 25, 50, 80 };


class kinectBasicApp : public AppBasic
{
public:
	void prepareSettings( Settings* settings );
	void setup();
	void mouseUp( MouseEvent event );
    void keyDown( KeyEvent event );
	void update();
	void draw();
    
    void startRecording(Recording::Color color = Recording::NONE);
    void startPlaybackOfCurrentRecording();
    
	Kinect mKinect;
	gl::Texture	mColorTexture, mDepthTexture;
    gl::GlslProg mShader;
    gl::GlslProg mLoopShader;
    gl::GlslProg mBGShader;
    
    ImageSourceRef mLastDepthImage;

    Recording r[NUM_RECORDINGS];
    int nextRecording;
    int currentRecording;
    
    VideoDelay * videoDelays;
    int numVideoDelays;
    
    VideoDelay bgDelay;
    
    BlobDetector mBlobDetector;
};

void kinectBasicApp::prepareSettings( Settings* settings )
{
	settings->setWindowSize( 640, 480 );
    settings->setFrameRate(30);
}


void kinectBasicApp::setup()
{
    Rand::randomize();
    
	console() << "There are " << Kinect::getNumDevices() << " Kinects connected." << std::endl;

	mKinect = Kinect( Kinect::Device() ); // the default Device implies the first Kinect connected
    
    nextRecording = 0;
    currentRecording = -1;
    
    mShader = gl::GlslProg( loadResource( RES_VERT_ID ), loadResource( RES_FRAG_ID ) );
    mLoopShader = gl::GlslProg( loadResource( RES_VERT_ID ), loadResource( RES_LOOP_FRAG_ID ) );
    mBGShader = gl::GlslProg( loadResource( RES_VERT_ID ), loadResource( RES_BG_FRAG_ID ) );
    
    numVideoDelays = NUM_VIDEO_DELAYS;
    if(numVideoDelays)
    {
        videoDelays = new VideoDelay[numVideoDelays];
        for(int i = 0; i < numVideoDelays; i++)
            videoDelays[i] = VideoDelay(getFrameRate()*VIDEO_DELAYS[i]);
    }
    
    bgDelay = VideoDelay(getFrameRate()*10, Vec3f(2,2,1), Vec2f(0,0));
    bgDelay.setThreshold(0.25);
    
    mBlobDetector.start();
}


void kinectBasicApp::mouseUp( MouseEvent event )
{
}


void kinectBasicApp::keyDown( KeyEvent event )
{
    switch(event.getChar())
    {
        case 'r':
            startRecording(Recording::RED);
            break;
        case 'g':
            startRecording(Recording::GREEN);
            break;
        case 'b':
            startRecording(Recording::BLUE);
            break;
            
        case 'p':
            startPlaybackOfCurrentRecording();
            break;
    }
}


void kinectBasicApp::update()
{	
	if( mKinect.checkNewDepthFrame() )
    {
        mLastDepthImage = mKinect.getDepthImage();
        
        for(int i = 0; i < NUM_RECORDINGS; i++)
        {
            if(r[i].isRecording())
            {
                r[i].addFrame(mLastDepthImage);
            }
        }
        
		mDepthTexture = mLastDepthImage;
    }
        
	if( mKinect.checkNewVideoFrame() )
		mColorTexture = mKinect.getVideoImage();
	
    if(mLastDepthImage)
    {
        for(int i = 0; i < numVideoDelays; i++)
        {
            videoDelays[i].update(mLastDepthImage);
        }
        
//        bgDelay.update(mLastDepthImage);
    }
    
    for(int i = 0; i < NUM_RECORDINGS; i++)
    {
        r[i].update();
    }
}

void kinectBasicApp::draw()
{
    gl::enableAlphaBlending();
    
	gl::clear( Color( 0, 0, 0 ) ); 
	gl::setMatricesWindow( getWindowWidth(), getWindowHeight() );
    
    
//    bgDelay.draw(mBGShader);
    
    
	if( mDepthTexture && mColorTexture )
    {
        mDepthTexture.bind(1);
        mColorTexture.bind(2);
		mShader.bind();
		mShader.uniform( "depthTex", (int)1 );
		mShader.uniform( "rgbTex", (int)2 );
		mShader.uniform( "threshold", 0.25f );
        mShader.uniform( "alphaFactor", 1.0f );
		gl::draw( mDepthTexture );
		mShader.unbind();
    }
    
    for(int i = 0; i < NUM_RECORDINGS; i++)
    {
        r[i].draw(mLoopShader);
    }
    
    for(int i = 0; i < numVideoDelays; i++)
    {
        videoDelays[i].draw(mLoopShader);
    }
    
    mBlobDetector.nextFrame(copyWindowSurface());
}


void kinectBasicApp::startRecording(Recording::Color color)
{
    if(currentRecording != -1)
    {
        r[currentRecording].stopRecording();
        r[currentRecording].startPlayback(true);
        currentRecording = -1;
    }
    
    r[nextRecording].clear();
    r[nextRecording].setColor(color);
    r[nextRecording].startRecording();    
    
    currentRecording = nextRecording;
    nextRecording = (nextRecording + 1) % NUM_RECORDINGS;
}


void kinectBasicApp::startPlaybackOfCurrentRecording()
{
    if(currentRecording != -1)
    {
        r[currentRecording].stopRecording();
        r[currentRecording].startPlayback(true);
        currentRecording = -1;
    }
}


CINDER_APP_BASIC( kinectBasicApp, RendererGl )
